function [y,alphas,sigmas,deltas]=batch_purelet_denoise(x,doPar,bkgCoords,useWTvar)
% Framewise PURE-LET denoising for 3D data `x'. 
% 
% This function is a wrapper that calls purelet_denoise iteratively for each 2D
% subarray of the 3D array `x' (data is sliced along the third dimension)
% 
% SYNTAX:
% 
% [y,alphas,sigmas,deltas]=batch_purelet_denoise(x)
% [y,alphas,sigmas,deltas]=batch_purelet_denoise(x,doPar)
% [y,alphas,sigmas,deltas]=batch_purelet_denoise(x,doPar,bkgCoords)
% 
% MANDATORY INPUTS:
% 
% x     : three dimensional grid data (array) e.g. a stack of 2D images (frames)
% 
% OPTIONAL INPUTS:
% 
% doPar     : scalar boolean, or empty matrix; when true, uses matlab parallel
%             processing toolbox for concurrent processing of each frame; else,
%             frames are processed in sequence; default is true if parallel
%             computing toolbox is installed, false otherwise
% 
% bkgCoords : cell array with two elements, or empty matrix; defines a
%             rectangular area for the calculation of noise parameters (gain,
%             offset, variance)
% 
%             Its elements are index arrays respectively into the first and
%             second dimension for each frame in `x' representing the data
%             subset for which noise parameters are calculated; whenever 
%             an element is the empty matrix, the whole extent of the
%             corresponding dimension will be used
% 
%             Example 1: 
% 
%             {1:50, 200:256} : defines the same rectangle which could be obtained
%             from the kth frame of x by calling directly x(1:50, 200:256,  k)
% 
%             Example 2: 
% 
%             {[], 200:256} : defines the same rectangle which could be obtained
%             from the kth frame of x by calling directly x(:, 200:256,  k)
% 
%             Example 3: 
% 
%             {1:50, []} : defines the same rectangle which could be obtained
%             from the kth frame of x by calling directly x(1:50, :,  k)
% 
%             Example 4: 
% 
%             {[], []} = defines the same rectangle which could be obtained
%             from the kth frame of x by calling directly x(:, :,  k) - i.e.,
%             the entire kth frame of x.
% 
%             Note that {[],[]} and {[]} are interpreted as representing the same
%             thing, i.e, the entire frame data.
% 
% useWTvar  : scalar boolean, default false; switch between AWGN variance
%             estimators: when true, use Donoho's robust AWGN variance estimator
%             in `wt_variance' routine; else, use the value returned from the
%             block processing routine `pawgn_noise_est'  
% 
% OUTPUTS:
% 
% y         : de-noised version of x
% 
% alphas,sigmas,deltas: AWGN parameters estimated for each frame of x
% 
% See also purelet_denoise, purelet_theta, pawgn_noise_est, wt_variance
% 
% References:
% 
% Luisier, F., T. Blu, and M. Unser. 2011. Image denoising in mixed
% Poisson-Gaussian noise. IEEE Trans. Image Process. 20:696?708.
% 
% C. M. Tigaret, K. Tsaneva-Atanasova, G. L. Collingridge, and J. R. Mellor.
% Wavelet transform-based de-noising for two-photon imaging of synaptic Ca2+
% transients. Biophysical Journal, 104(5):1006 ? 1017, 2013. 
% 
% Copyright 2011-2013 by Cezar M. Tigaret <Cezar.Tigaret@bristol.ac.uk>
  
%{
   ///***************************************************************************
   // *   Copyright (C) 2011-2013 by Cezar M. Tigaret <Cezar.Tigaret@bristol.ac.uk>  *
   // *                                                                         *
   // *                                                                         *
   // *   This program is free software; you can redistribute it and/or modify  *
   // *   it under the terms of the GNU General Public License as published by  *
   // *   the Free Software Foundation; either version 3 of the License, or     *
   // *   (at your option) any later version.                                   *
   // *                                                                         *
   // *   This program is distributed in the hope that it will be useful,       *
   // *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
   // *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
   // *   GNU General Public License for more details.                          *
   // *                                                                         *
   // *   You should have received a copy of the GNU General Public License     *
   // *   along with this program; if not, write to the                         *
   // *   Free Software Foundation, Inc.,                                       *
   // *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
   // ***************************************************************************/
%}


  sz = size(x);
  
  if(nargin<4 || isempty(useWTvar))
    useWTvar=false;
  end;
  
  if(nargin<3 || ~iscell(bkgCoords) || numel(bkgCoords)~=2)
    bkgCoords = {1:sz(1), 1:sz(2)};
  else
    if(isempty(bkgCoords{1}))
      bkgCoords{1} = 1:sz(1);
    end;
    if(isempty(bkgCoords{2}))
      bkgCoords{2} = 1:sz(2);
    end;
  end;
    
  if(nargin<2 || isempty(doPar))
    bb=which('matlabpool');
    if(isempty(bb))
      doPar=false;
    else
      doPar=true;
    end;
  end;
    
  if(ndims(x)==2 && ~isvector(x))
    NFrames=1;
  elseif(ndims(x)==3)
    NFrames = sz(3);
  else
    error(generatemsgid('BadArgument'), 'First argument must be a data volume (3D array)');
  end;
  
  y = NaN(sz);
  
  alphas = NaN(NFrames,1);
  sigmas = NaN(NFrames,1);
  deltas = NaN(NFrames,1);
  
  abort = false;

  if(doPar && NFrames>1)
    try
      matlabpool open;

      parfor k=1:NFrames
        fprintf('Frame %d...',k);

        data = x(:,:,k);
        
        bgCoords = bkgCoords;
        
        [alphas(k), sigmas(k), deltas(k)]=pawgn_noise_est(data(bgCoords{:}));
        
        if(useWTvar)
          sigmas(k) = wt_variance(data(bgCoords{:}));
        end;

        data = (data-deltas(k)) ./ alphas(k);

        y(:,:,k) = purelet_denoise(data, sigmas(k)/alphas(k)^2, false, false);

        y(:,:,k) = y(:,:,k) .* alphas(k) + deltas(k);
        
        fprintf(' done\n');
      end;

      matlabpool close;
    catch ME
      matlabpool close;
      for k=1:numel(ME.stack)
        fprintf('Error %s on line %s in file %s\n', ME.stack(k).name,ME.stack(k).line, ME.stack(k).file);        
      end;
      throw(ME);
    end;
  else
    try
      wbHandle=waitbar(0,'Please wait...',...
        'Name','Denoising... Press Cancel to stop. Do NOT close this window',...
        'CreateCancelBtn','setappdata(gcbf,''canceling'',1)');
      setappdata(wbHandle,'canceling',0);

      for k=1:NFrames
        fprintf('Frame %d\t',k);

        if(getappdata(wbHandle,'canceling'))
          abort = true;
          break;
        end;

        data = x(:,:,k);

        [alphas(k), sigmas(k), deltas(k)]=pawgn_noise_est(data(bkgCoords{:}));
        
        if(useWTvar)
          sigmas(k) = wt_variance(data(bkgCoords{:}));
        end;

        data = (data-deltas(k)) ./ alphas(k);

        y(:,:,k) = purelet_denoise(data, sigmas(k)/alphas(k)^2,false,true);

        y(:,:,k) = y(:,:,k) .* alphas(k) + deltas(k);

        waitbar(k/NFrames,wbHandle,sprintf('Frame %d of %d',k,NFrames));

      end;

      if(abort)
        y=x;
        alphas=[];
        sigmas=[];
        deltas=[];
      end;

      waitbar(1,wbHandle,'Done!');
      delete(wbHandle);
    catch ME
      delete(wbHandle)
      for k=1:numel(ME.stack)
        fprintf('Error %s on line %s in file %s\n', ME.stack(k).name,ME.stack(k).line, ME.stack(k).file);        
      end;
      throw(ME);
    end;

  end; % function batch_purelet_denoise
  

  
  
  
end % function batch_purelet_denoise