function xhat = purelet_denoise(x, sigma2, showProgress, doTimer)
%   
% PURE-LET de-noising for 2D data `x'
% 
% SYNTAX:
% 
% xhat = purelet_denoise(x, sigma2)
% xhat = purelet_denoise(x, sigma2, showProgress)
% xhat = purelet_denoise(x, sigma2, showProgress, doTimer)
% 
% MANDATORY INPUTS:
% 
% x      : two dimensional array (noisy data), double precision
% 
% sigma2 : estimated AWGN variance; scalar, double precision, or empty matrix;
%          when empty, the sigma2 is estimated from the detail coefficients of
%          the first level of wavelet transform (Donoho, 1993)
% 
% OPTIONAL INPUTS:
% 
% showProgress : scalar, boolean, or empty; default is true - show a progressbar
% 
% doTimer      : scalar, boolean, or empty; default is true - show elapsed
%                processing time
% 
% OUTPUT
% 
% xhat    - denoised version of `x'
% 
% See also batch_purelet_denoise, purelet_theta, pawgn_noise_est, wt_variance
% 
% References:
% 
% Luisier, F., T. Blu, and M. Unser. 2011. Image denoising in mixed
% Poisson-Gaussian noise. IEEE Trans. Image Process. 20:696?708.
% 
% Donoho, D. L. 1993. Nonlinear wavelet methods for recovery of signals,
% densities, and spectra from indirect and noisy data. In Proceedings of
% Symposia in Applied Mathematics. American Mathematical Society, Providence,
% RI. 173?205.
% 
% Copyright 2011-2013 by Cezar M. Tigaret <Cezar.Tigaret@bristol.ac.uk>
 
%{
   ///***************************************************************************
   // *   Copyright (C) 2011-2013 by Cezar M. Tigaret <Cezar.Tigaret@bristol.ac.uk>  *
   // *                                                                         *
   // *                                                                         *
   // *   This program is free software; you can redistribute it and/or modify  *
   // *   it under the terms of the GNU General Public License as published by  *
   // *   the Free Software Foundation; either version 3 of the License, or     *
   // *   (at your option) any later version.                                   *
   // *                                                                         *
   // *   This program is distributed in the hope that it will be useful,       *
   // *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
   // *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
   // *   GNU General Public License for more details.                          *
   // *                                                                         *
   // *   You should have received a copy of the GNU General Public License     *
   // *   along with this program; if not, write to the                         *
   // *   Free Software Foundation, Inc.,                                       *
   // *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
   // ***************************************************************************/
%}

  if(isvector(x) || ndims(x)>2)
    error(generatemsgid('BadArgument','Data must be a matrix'));
  end;
  
  if(nargin<4)
    doTimer = true;
  end;
  
  if(nargin<3 || isempty(showProgress))
    showProgress = true;
  end;
  
  if(nargin<2)
    sigma2 = 0;
  end;
  
  abort = false;
  
  % this will be initialized at first call only, as it's independent of the data
  persistent pStruct; 
  
  if(isempty(pStruct))
    pStruct.b0       = sqrt(2)./2;
    
    % quadrature mirror filter for low- and band-pass analysis -- Haar wavelet
    % (orthonormal)
    pStruct.qmf      = [1, 1] .* pStruct.b0;
    pStruct.fmq      = [1,-1] .* pStruct.b0;
    
    % implements
    %
    % t1 = x
    %            _                         _
    %           |     _           _ thetaP  |
    %           |    |    x        |        |
    %           |  - |-------------|        |
    %           |    | thetaR*t(w) |        |
    %           |    |_           _|        |
    %           |_                         _|
    % t2 = x * e                             
    pStruct.thetafun = @purelet_theta;
    
    pStruct.thetaR = 3;
    pStruct.thetaP = 8;
    pStruct.thetaK = 100;
    pStruct.ptitle = 'Denoising...';
   
  end;

  sx = size(x);
  
  N = numel(x);

  N3 = N*3;
  
  J = min(floor(log2(size(x))));
  
  NThetas=2*J;  
    
  Fy            = zeros(N,NThetas);
  yDelFy        = zeros(NThetas,1);
  yDiv_Fy_delFy = zeros(NThetas,1);
  
  % 1D frequency domains for X and Y dimensions
  freqx = 2*pi   .* linspace(0,1-1/sx(2),sx(2));
  z_x   = exp(1i .* freqx);

  freqy = 2*pi   .* linspace(0,1-1/sx(1),sx(1));
  z_y   = exp(1i .* freqy);
  
  if(doTimer)
    fprintf(pStruct.ptitle);
  end;
  
  if(showProgress)
    wbHandle=waitbar(0,'Please wait...',...
      'Name',[pStruct.ptitle,' Press Cancel to stop. Do NOT close this window'],...
      'CreateCancelBtn','setappdata(gcbf,''canceling'',1)');
    setappdata(wbHandle,'canceling',0);
  end;
  
  if(doTimer)
    tic
  end;
  
  % ----------------------------------------------------------------------------
  % forward transform and calculate PURE terms for each level
  %
  
  X = fft2(x);
  
  % low-pass reconstruction filter for Hadamard products using the
  % reconstruction matrix (DoR and QoR) starts here; for each level we build it
  % up recursively
  HH_ = ones(size(x));
    
  for j = 1:J  % decomposition level-wise iteration START
    
    if(showProgress && isvalid(wbHandle))
      if(getappdata(wbHandle,'canceling'))
        abort = true;
        break;
      end;
    end;
    
    j1 = j + J;
    
    beta_j = pStruct.b0 ^ j;
    
    %---------------------------------------------------------------------------
    % prepare diagonal terms
    
    diagsDR   = ones(1,N3).*(2^(-2*j));
    diagsQR   = ones(1,N3).*(2^(-2*j));
    diagsDoDR = diagsDR .* (sqrt(2)/20);
    diagsQoQR = -diagsDoDR;
    diagsDoQR = diagsDR./2;
    
    %-------------------------------------
    % calculate transform coefficients
    
    % prepare transfer functions
    
    zexp = 2^(j-1); % dilation operator for the current level `j'

    % frequency domains for analysis at current level
    zx_ = z_x    .^ -zexp; % row vector 
    zy_ = z_y(:) .^ -zexp; % column vector

    % frequency domains for synthesis at current level -- needed for DoR and
    % QoR Hadamard product terms
    zx  = z_x    .^ zexp; % row vector
    zy  = z_y(:) .^ zexp; % column vector

    % 1D frequency responses for forward transform (analysis)
    
    %  low-pass
    Hx = polyval(pStruct.qmf,zx_); % horizontal: row vector  
    Hy = polyval(pStruct.qmf,zy_); % vertical:   col vector 

    %  high-pass
    Gx = polyval(pStruct.fmq,zx_); % horizontal
    Gy = polyval(pStruct.fmq,zy_); % vertical

    % 1D frequency responses for DoR and QoR forward transform (Hadamard
    % products) and for reconstruction (inverse transform)
    % divide by 2 for energy conservation
    
    %   low-pass
    Hx_ = polyval(pStruct.qmf,zx)./2; % horizontal: row vector  
    Hy_ = polyval(pStruct.qmf,zy)./2; % vertical:   col vector 
    %   high-pass
    Gx_ = polyval(pStruct.fmq,zx)./2; % horizontal: row vector 
    Gy_ = polyval(pStruct.fmq,zy)./2; % vertical:   col vector 
    
    HH = Hy * Hx;
    
    GH = Gy * Hx;
    HG = Hy * Gx;
    GG = Gy * Gx;

    %  HH_ - computed at end of iteration
    GH_ = Gy_ * Hx_;
    HG_ = Hy_ * Gx_;
    GG_ = Gy_ * Gx_;
    
    % for reconstruction filters in the DoR and QoR Hadamard products
    
    ghhh = ifft2(GH_ .* HH_);
    hghh = ifft2(HG_ .* HH_);
    gghh = ifft2(GG_ .* HH_);
    
    xa = X .* HH; % => approximation coefficients at current level

    xh = real(ifft2(X .* GH)); % => horizontal detail coefficients at current level
    xv = real(ifft2(X .* HG)); % => vertical   detail coefficients at current level
    xd = real(ifft2(X .* GG)); % => diagonal   detail coefficients at current level

    xdorh = real(ifft2(X .* fft2(ifft2(GH) .* ghhh)));
    xdorv = real(ifft2(X .* fft2(ifft2(HG) .* hghh)));
    xdord = real(ifft2(X .* fft2(ifft2(GG) .* gghh)));
    
    % no group delay compensation for Haar wavelets
    xqh = real(ifft2(xa));
    xqv = xqh;
    xqd = xqh;

    xqorh = real(ifft2(X .* fft2(ifft2(HH) .* ghhh)));
    xqorv = real(ifft2(X .* fft2(ifft2(HH) .* hghh)));
    xqord = real(ifft2(X .* fft2(ifft2(HH) .* gghh)));

    d = [xh(:);xv(:);xd(:)];
    
    if(isempty(sigma2) && j==1)
      sigma2 = (median(abs(d(:)))/0.6745)^2;
    end;
    
    v = [xqh(:);xqv(:);xqd(:)];

    xdor = [xdorh(:);xdorv(:);xdord(:)];
    xqor = [xqorh(:);xqorv(:);xqord(:)];

    [t, dtdx, dtdy, dtdxdy, dtdx2, dtdy2] = ...
      pStruct.thetafun(d, v, sigma2, beta_j, pStruct.thetaR, pStruct.thetaP, pStruct.thetaK);

    %---------------------------------------------------------------------------
    % Calculate the terms in the PURE equation
    %
    % the Poisson part: y*delF(y) ( 2nd term of Taylor approximation of F_(y) )
    %
    % the AWGN part of PURE ( first 2 terms of Taylor approximation of F_(y) )
    %
    % y * div{F(y)-delF(y)} = y * div{F(y)} - y * div{delF(y)} , where:
    %
    % y * div{F(y)}    = diag{DR}     * dt/du + diag{QR}  * dt/dv
    %
    % y * div{delF(y)} = diag{(DoD)R} * d2t/du2 + diag{(QoQ)R} * d2t/dv2 - 2 * diag{(DoQ)R} * dt/(du*dv)
    %
    yDelFy(j)  = ( dtdx(:,1).' * xdor(:) + dtdy(:,1).' * xqor(:) ) ./ N3;

    yDelFy(j1) = ( dtdx(:,2).' * xdor(:) + dtdy(:,1).' * xqor(:) ) ./ N3;

    yDiv_Fy_delFy(j)   = (diagsDR    * dtdx(:,1)    + diagsQR   * dtdy(:,1)  - ...
                          (diagsDoDR * dtdx2(:,1)   + diagsQoQR * dtdy2(:,1) - ...
                           2 * diagsDoQR * dtdxdy(:,1))) ./ N3;

    yDiv_Fy_delFy(j1)  = (diagsDR    * dtdx(:,2)    + diagsQR   * dtdy(:,2)  - ...
                          (diagsDoDR * dtdx2(:,2)   + diagsQoQR * dtdy2(:,2) - ...
                           2 * diagsDoQR * dtdxdy(:,2))) ./ N3;

    %---------------------------------------------------------------------------
    % Then reconstruct the subband data (Fy) from thresholded coefficients from
    % the j level only (i.e., everything else set ot zero)
    %
    %   theta 1 and theta 2, separately
    %
    xt1h = fft2(reshape(t(1:N,         1), sx)) .* GH_; %   horizontal orientation
    xt1v = fft2(reshape(t((N+1):2*N,   1), sx)) .* HG_; %   vertical orientation
    xt1d = fft2(reshape(t((2*N+1):3*N, 1), sx)) .* GG_; %   diagonal orientation

    xt2h = fft2(reshape(t(1:N,         2), sx)) .* GH_; %   horizontal orientation
    xt2v = fft2(reshape(t((N+1):2*N,   2), sx)) .* HG_; %   vertical orientation
    xt2d = fft2(reshape(t((2*N+1):3*N, 2), sx)) .* GG_; %   diagonal orientation

    fy_t1 = real(ifft2((xt1h + xt1v + xt1d).* HH_));    % => Fy with theta 1
    fy_t2 = real(ifft2((xt2h + xt2v + xt2d).* HH_));    % => Fy with theta 2

    Fy(:,j)  = fy_t1(:);
    Fy(:,j1) = fy_t2(:);


    %---------------------------------------------------------------------------
    % go on to the next level
    
    X = xa; % stays in the Fourier domain
    
    % low-pass reconstruction filter for Hadamard products using the
    % reconstruction matrix (DoR and QoR) -- prepare for next level
    HH_ = HH_ .* (Hy_ * Hx_);
        
    if(showProgress && isvalid (wbHandle))
      waitbar(j/J,wbHandle,sprintf('Level %d of %d',j,J));
    end;
    
  end; % decomposition level-wise iteration END
  
  % close progress bar if shown
  if(showProgress && isvalid (wbHandle))
    waitbar(1,wbHandle,'Done!');
    delete(wbHandle);
  end;
  
  if(abort)
    xhat=[];
    return;
  end;

  %------------------------------------------------------------------
  % reconstruct the low-pass (approximation) data in the image domain
  %------------------------------------------------------------------
  Fy_lp = real(ifft2(X .* HH_));

  %-----------------------------------------------------------------------------
  % finally compute the PURE vector `c', and solve for the linear parameters `a'
  %-----------------------------------------------------------------------------

  yHi = x(:) - Fy_lp(:); % this is faster, as we already have got these terms

  t_ = false(J,1);
  t_(1:3) = true;
  
  endx=repmat(t_,2,1);
  
  M = (Fy.' * Fy)./N;
  
  yFy  = Fy.' * yHi(:)./N;
  c = (yFy - yDelFy - sigma2 * yDiv_Fy_delFy);


  a=pinv(M)*c;
    
  aa=a;

  aa(endx)=0;
  
  xhat = reshape(Fy*aa,sx)+Fy_lp;

  xhat(abs(xhat)<eps)=0; % keep it clean

  xhat(xhat<0)=0;        % we don't do negative image data...

  if(doTimer)
    toc
  end;
  
end % function purelet_denoise