function [t, dt_dx, dt_dy, dt_dxdy, dt2_dx2, dt2_dy2, atwsqr]=...
            purelet_theta(x,y,v,b,r,p,k,debugOn)

% SYNTAX:
%
%            [t, dt_dx, dt_dy, dt_dxdy, dt2_dx2, dt2_dy2]=...
%             purelet_theta(x,y,v,b,r,p,k,debugOn)
%
% INPUTS:
% x, y       = wavelet and scaling (or GDC) coefficients, respectively
% v          = variance of AWGN
% b          = scale-dependent factor (2^(-j/2))
% r,p,k      + theta function parameters (see Luisier et al 2011)
% 
% implements
%
% t1 = x
%             _               _
%            |     _      _ p  |
%            |    |    x   |   |
%            |  - |--------|   |
%            |    | r*t(w) |   |
%            |    |_      _|   |
%            |_               _|
% t2 = x * e
%
% where t(w) = sqrt(b * w + v), 
%
% with w = y*tanh(k*y), k=100, a differentiable approximation of abs(y)
%   
% and typically with r = 3 and p = 8
%   
% the corresponding derivatives of the theta functions will be taken with
% respect to w (which aproximates |y|) and NOT to y itself !!!
%
% 
% because the image data is a function that can take only positve integer
% numbers, imposing unity as a lower bound for denominators when calculating
% the derivatives does not bring a great loss of generality, while avoiding
% unstable results (i.e., NaNs) especially for very low intensity signals
%   
% Reference:
% 
% Luisier, F., T. Blu, and M. Unser. 2011. Image denoising in mixed
% Poisson-Gaussian noise. IEEE Trans. Image Process. 20:696?708.
% 
% Copyright 2011-2013 by Cezar M. Tigaret <Cezar.Tigaret@bristol.ac.uk>

%{
   ///***************************************************************************
   // *   Copyright (C) 2011-2013 by Cezar M. Tigaret <Cezar.Tigaret@bristol.ac.uk>  *
   // *                                                                         *
   // *                                                                         *
   // *   This program is free software; you can redistribute it and/or modify  *
   // *   it under the terms of the GNU General Public License as published by  *
   // *   the Free Software Foundation; either version 3 of the License, or     *
   // *   (at your option) any later version.                                   *
   // *                                                                         *
   // *   This program is distributed in the hope that it will be useful,       *
   // *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
   // *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
   // *   GNU General Public License for more details.                          *
   // *                                                                         *
   // *   You should have received a copy of the GNU General Public License     *
   // *   along with this program; if not, write to the                         *
   // *   Free Software Foundation, Inc.,                                       *
   // *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
   // ***************************************************************************/
%}

    tol=1e-6;

    if(~exist('debugOn','var'))
      debugOn=false;
    end;

    
    if(debugOn)
      if(any(any(isnan(x))))
        warning(generatemsgid('BadArgument'),'NaN in x')
      end;

      if(any(any(isinf(x))))
        warning(generatemsgid('BadArgument'),'Inf in x')
      end;

      if(any(any(isnan(y))))
        warning(generatemsgid('BadArgument'),'NaN in y')
      end;

      if(any(any(isinf(y))))
        warning(generatemsgid('BadArgument'),'Inf in y')
      end;

    end;
    
    if(~all(size(x)==size(y)))
      error(generatemsgid('BadArgument'),'x and y must have the same size');
    end;

    ky = k.*y;
    
    w  = y .* tanh(ky);    % the smooth approximation of abs(y) = |y|
    
    tw = b .* (w + v) + tol; % the scale-dependent variance estimate: t(w) = beta * |y| + sigma^2
    
    twsqr  = sqrt(tw);     

    atwsqr = r*twsqr;   
    
    atwsqr(atwsqr==0) = tol;
                            %                             x
    pexp = x./atwsqr;       % the exponent in gT: ---------------
                            %                       r * sqrt(t(w))
                            %
    if(any(isnan(pexp)))
      pexp(isnan(pexp))=0;
      if(debugOn)
        warning(generatemsgid('RuntimeError'),'NaN in pexp');
      end;
    end;

    if(any(isinf(pexp)))
      pexp(isinf(pexp)) = 0;
      if(debugOn)
        warning(generatemsgid('RuntimeError'),'Inf in pexp');
      end;
    end;
    
    gTe  = pexp.^p;         % to the power of p
    
    gT = exp(-gTe);         % 
    
    pgTe  = gTe.* p;
    
    x(x==0) = tol;
    
    bp2   = b*p/2;
    b4    = b^2/4;
    gT_tw = gT./tw;
    gT_x  = gT./x;
    x_tw  = x./tw;
    

    if(any(isnan(gT_tw)))
      gT_tw(isnan(gT_tw))=0;
      if(debugOn)
        warning(generatemsgid('RuntimeError'),'NaN in gT/tw');
      end;
    end;

    if(any(isinf(gT_tw)))
      gT_tw(isinf(gT_tw))=0;
      if(debugOn)
        warning(generatemsgid('RuntimeError'),'Inf in gT/tw');
      end;
    end;

    if(any(isnan(gT_x)))
      gT_x(isnan(gT_x))=0;
      if(debugOn)
        warning(generatemsgid('RuntimeError'),'NaN in gT/x');
      end;
    end;

    if(any(isinf(gT_x)))
      gT_x(isinf(gT_x))=0;
      if(debugOn)
        warning(generatemsgid('RuntimeError'),'Inf in gT/x');
      end;
    end;

    if(any(isnan(x_tw)))
      x_tw(isnan(x_tw))=0;
      if(debugOn)
        warning(generatemsgid('RuntimeError'),'NaN in x/tw');
      end;
    end;

    if(any(isinf(x_tw)))
      x_tw(isinf(x_tw))=0;
      if(debugOn)
        warning(generatemsgid('RuntimeError'),'Inf in x/tw');
      end;
    end;

    
    %
    % threshold functions: theta1 and theta2
    %
    
    t=[x, x.*gT];
    
    
    if(nargout==1)
      return;
    end;
    
    %
    % prepare to calculate derivatives
    %
    
    
    %
    % derivatives for theta1
    %

    %   first order
    d1t1_dx   = ones(size(x));  % in x
    d1t1_dy   = zeros(size(x)); % in y
    d1t1_dxdy = zeros(size(x)); % in x and y

    %   second order
    d2t1_dx   = zeros(size(x)); % in x
    d2t1_dy   = zeros(size(x)); % in y

    %
    % derivatives for theta 2
    %

    %   first order
    d1t2_dx = (1 - pgTe) .* gT; % in x
    d1t2_dy = bp2 .* gT_tw .* gTe .* x; % in w (i.e. the approximation of abs(y)
    d1t2_dxdy = bp2 .* gTe .* gT_tw .* (1 + p - pgTe); % in x and in w
    
    %
    %   second order
    d2t2_dx  = gT_x .* pgTe .* (pgTe - p - 1); % in x
    d2t2_dy  = b4 .* gT_tw .* x_tw .* pgTe .* (pgTe - p - 4); % in w
    
    
    if(any(isnan(d1t2_dx)))
      d1t2_dx(isnan(d1t2_dx))=0;
      if(debugOn)
        warning(generatemsgid('RuntimeError'),'NaN in d1t2_dx');
      end;
    end;
    
    if(any(isinf(d1t2_dx)))
      d1t2_dx(isinf(d1t2_dx))=0;
      if(debugOn)
        warning(generatemsgid('RuntimeError'),'Inf in d1t2_dx');
      end;
    end;
    
    if(any(isnan(d1t2_dy)))
      d1t2_dy(isnan(d1t2_dy))=0;
      if(debugOn)
        warning(generatemsgid('RuntimeError'),'NaN in d1t2_dy');
      end;
    end;
    
    if(any(isinf(d1t2_dy)))
      d1t2_dy(isinf(d1t2_dy))=0;
      if(debugOn)
        warning(generatemsgid('RuntimeError'),'Inf in d1t2_dy');
      end;
    end;
    
    if(any(isnan(d1t2_dxdy)))
      d1t2_dxdy(isnan(d1t2_dxdy))=0;
      if(debugOn)
        warning(generatemsgid('RuntimeError'),'NaN in d1t2_dxdy');
      end;
    end;
    
    if(any(isinf(d1t2_dxdy)))
      d1t2_dxdy(isinf(d1t2_dxdy))=0;
      if(debugOn)
        warning(generatemsgid('RuntimeError'),'Inf in d1t2_dxdy');
      end;
    end;
    
    if(any(isnan(d2t2_dx)))
      d2t2_dx(isnan(d2t2_dx))=0;
      if(debugOn)
        warning(generatemsgid('RuntimeError'),'NaN in d2t2_dx');
      end;
    end;
    
    if(any(isinf(d2t2_dx)))
      d2t2_dx(isinf(d2t2_dx))=0;
      if(debugOn)
        warning(generatemsgid('RuntimeError'),'Inf in d2t2_dx');
      end;
    end;
    
    if(any(isnan(d2t2_dy)))
      d2t2_dy(isnan(d2t2_dy))=0;
      if(debugOn)
        warning(generatemsgid('RuntimeError'),'NaN in d2t2_dy');
      end;
    end;
    
    if(any(isinf(d2t2_dy)))
      d2t2_dy(isinf(d2t2_dy))=0;
      if(debugOn)
        warning(generatemsgid('RuntimeError'),'Inf in d2t2_dy');
      end;
    end;
    
    dt_dx   = [d1t1_dx(:), d1t2_dx(:)];
    dt_dy   = [d1t1_dy(:), d1t2_dy(:)];
    
    dt2_dx2 = [d2t1_dx(:), d2t2_dx(:)];
    dt2_dy2 = [d2t1_dy(:), d2t2_dy(:)];
    
    dt_dxdy = [d1t1_dxdy(:), d1t2_dxdy(:)];
      
    
  end % function purelet_theta
  

  