function y = wt_variance(x)
%   
% Estimates AWGN variance in `x' from the median absolute deviation of the
% wavelet coefficients at highest resolution (Haar domain), according to the
% formula in :
% Donoho, D. L. 1993. Nonlinear wavelet methods for recovery of signals,
% densities, and spectra from indirect and noisy data. In Proceedings of
% Symposia in Applied Mathematics. American Mathematical Society, Providence,
% RI. 173?205. 
% 
% SYNTAX:
% 
% y = wt_variance(x)
% 
% INPUT:
% 
% x  : matrix data
% 
% OUTPUT:
% 
% y  : estimated variance
% 
% Copyright 2011-2013 by Cezar M. Tigaret <Cezar.Tigaret@bristol.ac.uk>
% 

%{
///***************************************************************************
// *   Copyright (C) 2011-2013 by Cezar M. Tigaret <Cezar.Tigaret@bristol.ac.uk>  *
// *                                                                         *
// *                                                                         *
// *   This program is free software; you can redistribute it and/or modify  *
// *   it under the terms of the GNU General Public License as published by  *
// *   the Free Software Foundation; either version 3 of the License, or     *
// *   (at your option) any later version.                                   *
// *                                                                         *
// *   This program is distributed in the hope that it will be useful,       *
// *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
// *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
// *   GNU General Public License for more details.                          *
// *                                                                         *
// *   You should have received a copy of the GNU General Public License     *
// *   along with this program; if not, write to the                         *
// *   Free Software Foundation, Inc.,                                       *
// *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 // ***************************************************************************/
%}


  
  if(isvector(x) || ndims(x)>2)
    error(generatemsgid('BadArgument','Data must be a matrix'));
  end;
  
  sx=size(x);
  
  b0  = sqrt(2)./2;
  qmf = [1, 1] .* b0;
  fmq = [1,-1] .* b0;
  
  % 1D frequency domains for X and Y dimensions
  freqx = 2*pi   .* linspace(0,1-1/sx(2),sx(2));
  zx_   = exp(1i .* freqx) .^ -1;

  freqy = 2*pi   .* linspace(0,1-1/sx(1),sx(1));
  zy_   = exp(1i .* freqy) .^ -1;
  
  
  X = fft2(x);
  
  
  Hx = polyval(qmf,zx_);    % horizontal: row vector  
  Hy = polyval(qmf,zy_(:)); % vertical:   col vector 

  %  high-pass
  Gx = polyval(fmq,zx_);    % horizontal
  Gy = polyval(fmq,zy_(:)); % vertical

  GH = Gy * Hx;
  HG = Hy * Gx;
  GG = Gy * Gx;

  xh = real(ifft2(X .* GH)); % => horizontal detail coefficients at current level
  xv = real(ifft2(X .* HG)); % => vertical   detail coefficients at current level
  xd = real(ifft2(X .* GG)); % => diagonal   detail coefficients at current level
  
  y = (median(abs([xh(:);xv(:);xd(:)]))/0.6745)^2;
  
end % function wt_variance